# This program correlates weather data (temperature,pressure) against the rate of cosmic ray events.
#
# It reads two files downloaded from the HiSPARC public database (e.g. https://data.hisparc.nl/show/stations_by_country/).
#    One file contains weather data for the selected period, the other contains cosmic ray events.
#
#    Run this program with 
#        python  weather_events.py   <weather-file> <event-summary-file>
#    where the actual files names should be substituted for the place-holders delimited by <>.
#
#    The files should cover the same time period - though not necessarily having the same sampling rate
#    (e.g. weather data is sampled every 5 minutes, but CR events arrive randomly with many every 5 minutes).
#
# Both files are in the standard HiSPARC data-download format:
#    a) An explantory or "header" section describing the data in the rest of the file.
#       Each line in the header starts with a hash (#) and should be ignored by programs reading the file.
#       The contents of the header are self explantory, containing the time interval for the data
#       and a description of the columns in the rest of the file.
#    b) Lines after the header are all in multi-column format, 
#       with each column separated from its neighbours by a "tab" character.
# Further description of meaning of data in each of these columns can be found in the
# documentation on the HiSPARC website https://docs.hisparc.nl.
#
# The program extracts relevant columns of data from each of the files.
#
# The output from this program consists of a tab-separated values file with columns
# 1) time       denoting the time at the end of a period in which weather data has been averages
#               and the number of events counted. The time intervals will generally be uniform.
#               Units: seconds since the "epoch" - i.e. the start of 1970 - corresponding
#               to time values logged from GPS.
# 2) Year       Integer - 4 digits
# 3) Month      In the range 0-11
# 4) DayofMonth In the range 0-30
# 5) HourOfDay  In the range 0-23. Useful for day/night comparisons. 
# 6) average    average temperature over the interval - units Deg C 
# 7) average    average barometric pressure over the time interval  units milliBar
# 8) eventRate  the number of events logged in that time interval divided by the interval.
#
#

import sys          # Python standard library - lets us interact with the operating system
                    # for example, to get parameters from the command line used to start the program.
import re           # Python standard library - regular expressions.    
import math         # Python standard library - all the usual mathematical functions (sin, cos etc.).
import datetime     # Standard conversions between different date and time formats.
import matplotlib as plot   # Plotting library
import matplotlib.pyplot as plt   # Plotting library
debug = 0           # Values above zero print additional information on standard output,
                    # to allow closer analysis of behaviour and help diagnose errors.

#=============================================================================================================
#
#   This group of sub-programs represents the Level 2 decomposition of dataInput(): 
#   i.e. those sub-programs called directly from the dataInput() sub-program.
#
#   It includes:
#     readCommandLine()                         get parameters from the command line specifying two files.
#     readWeatherData(weatherFIle):             read a file containing the weather data.
#     readEventSummaryFile(eventSummaryFile):   read a file containing event data.

def readCommandLine():
    """Expects the program command line to contain paths to two file names.
    
       Return two values (both character strings) as a two valued list
       The first element is the name of the weather data file,
       The second element is the event data file.
       
       IF the command lines does NOT contain at least two items then declare an error.
    """
    
    cmdLineArgs = sys.argv      # This is just a list that is present in Python
                                # if you "import sys".
    
    nargs           = len(cmdLineArgs)
    if(nargs>=2):
        weatherFile     = cmdLineArgs[1]
    else:
        weatherFile = ""
        print "No weather data file name supplied!"
        
    if(nargs>=3):
        eventSummaryFile= cmdLineArgs[2]
    else:
        eventSummaryFile = ""
        print "No event data file name supplied!"    
        
    return (weatherFile,eventSummaryFile)


def readFileByColumns(dataFile,comment,sep,names):
    """
    Read a file containing columns of data deliniated by a designated character.
    
    Input:
        dataFile    The file system path to a file containing the data.
        comment     The character that is used to introduce comment lines.
                    (If the first character matches this, then ignore the line.)
        sep         The character used to separate columns in the data file
                    (this is normally a tab character - or sometimes a comma).
        names       A list of character string to be used to name the columns.
    
    Output:         Return value - a Python dictionary in which
                    the column names are used as keys and the associated values
                    are Python arrays containing the data associated with the 
                    named column.
    """
    
    fp = open(dataFile,"r")
    
    columns = {}
    
    for name in names:
        columns[name] = []
    
    count = 0
    for line in fp:
        line = line.lstrip()
        
        if(line[0] != comment):
            
            count = count+1
            line = line.rstrip()        
            
            words = line.split(sep)
        
            nwords = len(words)
            
            for iw in range(nwords):
                columns[names[iw]].append(words[iw])
    
    return columns

def readWeatherData(weatherFile):
    """Read the weather data file,
       and convert it to an internal format that is useful for further processing.
       
       The data files should be in the standard "download" format of HiSPARC,
       which writes an "ASCII" file (i.e. containing characters and viewable in a text editor
       with each line corresponding to a "record"). The lines are in two groups
         a) An explantory or "header" section describing the data in the rest of the file.
            Each line in the header starts with a hash (#) and should be ignored by programs reading the file.
            The contents of the header are self explantory, containing the time interval covered by the data
            and a description of the columns in the rest of the file.
         b) Lines after the header are all in multi-column format, 
            with each column separated from its neighbours by a "tab" character.
    
        Return a PANDAS "dataframe" which is a box containing a table (a bit like a spreadsheet)
        composed of columns (each of which is individually labelled).
    """
    
    # Note that the raw data in which each record read from the file is just a string of characters,
    # whereas for our further processing we need the data to be stored in one of the Python number formats
    # (e.g. integers, floats, doubles etc.). I.e. the six character string "1021.4" must become the number 1021.4.
    #
    #
    
    # This returns a dictionary, weatherData, where weatherData[<colname>] is an Python list of the data in that column.
    
    weatherData = readFileByColumns( weatherFile,
                                        comment='#',
                                        sep='\t',
                                        names=["date","time","timestamp",
                                        "temp_inside","temp_outside",
                                        "humidity_inside","humidity_outside",
                                        "pressure",
                                        "wind_direction",
                                        "wind_speed",
                                        "solar_radiation",
                                        "uv_index",
                                        "evapotranspiration",
                                        "rain_rate",
                                        "heat_index",
                                        "dew_point",
                                        "wind_chill"]
                                        )
    
    
    return weatherData                 # Further processing will be done in another function.
    
    
def readEventSummaryFile(eventSummaryFile):
    """Read the event summary data file,
       and convert it to an internal format that is useful for further processing.
       
       The data files is in the standard "download" format of HiSPARC,
       which writes an "ASCII" file (i.e. containing characters and viewable in a text editor
       with each line corresponding to a "record"). The lines are in two groups
         a) An explantory or "header" section describing the data in the rest of the file.
            Each line in the header starts with a hash (#) and should be ignored by programs reading the file.
            The contents of the header are self explantory, containing the time interval covered by the data
            and a description of the columns in the rest of the file.
         b) Lines after the header are all in multi-column format, 
            with each column separated from its neighbours by a "tab" character.
         c) Note that there are NO column titles in these files.
            We have to supply them explicitly while reading the file.
       
       For this particular file there are 23 columns of which will we eventually need only the time at which an event was logged.

    """
    
    # This returns a dictionary, eventData, where eventData[<colname>] is an Python list of the data in that column.
    eventData = readFileByColumns(eventSummaryFile,
                                comment='#',
                                sep='\t',
                                names=["date","time","timestamp","nanoseconds",
                                "pulseheights1","pulseheights2","pulseheights3","pulseheights4", 
                                "integrals1","integrals2","integrals3","integrals4",
                                "mips1","mips2","mips3","mips4",
                                "arrival_time1","arrival_time2","arrival_time3","arrival_time4",
                                "trigger_time",
                                "zenith","azimith"]
                                )
    
    
    return eventData

#============================================================================================================================

#=============================================================================================================
#
# These sub-programs define a Level 2 decomposition with sub-programs called directly from processData():
#

def aveWeatherData(averagingInterval, weatherData):
    """
       For this particular file there are 17 columns of which we need only three,
       those which contain the time at which data was logged, and the pressure and the temperature logged at that time.
        
       Input:
        averagingInterval   (Integer)           a value in seconds defining the period of averaging.
        weatherData         (PANDAS dataframe)  a data structure containing a table of weather data.
       
       Hence, this function must return three lists which will respectively contain
        timeZero       The time at which the time series starts (usually a little BEFORE
                       the time of the first data point).
        averagingTimes the nominal time at which the averaged data are to be located.
        pressureAve    the average barometric pressure over this time interval. A floating point number.
        tempAve        the average temperature over the time interval. A floating point number.
        
       The indexing of the data is such that the values pressureAve[i],tempAve[i]
                       will represent data logged at the time in averagingTimes[i].
    """
    
    (timeZero,averagingTimes,tempAve)    = processToAverages(weatherData,averagingInterval,"temp_outside")
    (timeZero,averagingTimes,pressureAve)= processToAverages(weatherData,averagingInterval,"pressure")
    
    
    return (timeZero,averagingTimes,tempAve,pressureAve)


def processToAverages(dataStore, averagingInterval, columnName):
    """ Take raw data logged on a frequent and even time interval and produce an array
        of data averaged over a longer interval.
        
        Input:
            dataStore           A collection of data containing multiple data time-series
                                This is represented as a Python dictionary, where
                                dataStore[<seriesName>] is an array represented one
                                type of data from HiSPARC (e.g. "temperature").
                                * One of the keys MUST be "timestamp" and must identify
                                  a data array of GPS timestamps (i.e. seconds with respect
                                  to the "epoch" of the start of the year 1970).
                                * The input data must be logged at even intervals.
                                * All the arrays in dataStore MUST have the same length.
            averagingInterval   The number of seconds over which data is to be averaged.
            columnName          The name of the time series (which will appear as a 
                                dictionary key in dataStore{}) to which the averaging
                                process will be applied.
        
        This process needs to return two arrays of equal length:
        a) A list of times (separated by averagingInterval)
        b) A list of the averages of the data identified by columnName
        
        What we therefore need to do is:
        1) Find the total length of the data 
        2) Get data arrays corresponding to the timestamp column and the data column identified by the columnName parameter.
        3) Work through the dataFrame line by line, checking the timestamp column to see in which hourly slot it sits.
        4) Add up all the data values inside one hourly slot, average them, 
           and write to the appropriate slot in an output array.
    """
    
    numPoints     = len( dataStore[columnName] )    # All columns should have the same length,
                                                    # so it should not matter which we use.
    timestamp     = dataStore["timestamp"]
    dataToAverage = dataStore[columnName]
    
    timeZero      = int(timestamp[0])-1             # This is a bit a bit cludgy - but leave it for now.
                                                    # I am building into the code an assumption that the
                                                    # real start of the data logging is 1 second before the first
                                                    # item of logged data. (This is the way HiSPARC weather data is presented.)
                                                    # HOWEVER, it is not good practice, because if HiSPARC logging practice changes
                                                    # The averaging algorithm here will go wrong.    
    
    if( debug >= 1):
        print "Number of points for ",columnName,":",numPoints, timeZero
    
    # We need arrays to hold output data.
    # It would have been possible to work out the expected number of time points in these
    # arrays by looking at the difference between the time of the first and last logged data point,
    # then dividing by the averaging interval. This would probably be more resource efficient.
    # Doing it this way - creating an empty array of zero length and then tacking data on the end
    # item by item, is certainly LESS efficient, but it makes it easier to follow the algorithm.
    # SO, I do it this way to start with, and might improve it later on if time usage become inconvenient.
    # My advice is: get something working quickly in an OBVIOUSLY RIGHT WAY, then improve later.
    times   = []    # Define an empty array of times.
    aveData = []    # Define an empty array to hold the output data.
    
    sumData  = 0.0  # Accumulate the sum of all items in the averaging interval.
    nData    = 0    # Number of items in the interval.
    nAvePoint= 0    # Number of output points created.
    
    for ip in range(numPoints):
        
        # Determine the time at which the averaging interval is to end.
        aveTime= timeZero+(nAvePoint+1)*averagingInterval        
        
        if( int(timestamp[ip]) <= aveTime ):
            nData   = nData   + 1
            sumData = sumData + float(dataToAverage[ip])    # There is an implicit assumption in this averaging
                                                            # calculation that the input points are evenly distributed
                                                            # in time. A better algorithm could take account of 
                                                            # non-uniform data logging intervals by summing the
                                                            # product of data*latest-time-interval,
                                                            # then dividing by the averaging interval to get the average.
                                                            # However, dealing with time intervals at the begging and end
                                                            # of an averaging interval is an potential error trap.
                                                            # So we keep it simple for now.
        else:
            # We have reach the end of one of the averaging intervals,
            # so calculate the average and add it to the end of the output arrays.
            nAvePoint = nAvePoint+1
            aveData.append(sumData/nData)
            times.append( aveTime )    # Assemble array of datetime objects.
            
            nData     = 0              # Reset the averaging accumulators for the next time interval 
            sumData   = 0.0
            
    if( debug >= 2):
        print ip,timestamp[ip],  datetime.datetime.utcfromtimestamp(int(timestamp[ip]) ), aveTime
    
    return (timeZero, times, aveData)


def getAverageEventRate(timeZero,averagingTimes,eventData):
    """ This process needs to return a count of the events that took place in that intervals
        designated by the input array averagingTimes.
        
        Note that this algorithm is safe if the weather data times cover periods
        before the first cosmic ray events are recorded. 
        We just get zero event rate in the output file.
        We also stop calculating averages if we go beyond the last weather data time.
        If the cosmic ray logging does not continue to the end of the weather data times,
        we add extra zero-rate data to the output array.
    """
    
    eventTimes = eventData["timestamp"]
    numevents = len(eventTimes)
    
    if( debug > 0):
        print "number of CR events:",numevents
        print "number of ave times",len(averagingTimes)
    
    eventRate = []
    
    count = 0
    at    = 0
    interval = averagingTimes[0]-timeZero
    

    for ip in range(numevents):
        #print count,at,averagingTimes[at] ,int(eventTimes[ip])
        if( int(eventTimes[ip])< averagingTimes[at]):
            count = count + 1
        else:
            rate = 3600.0*float(count)/interval
            eventRate.append(rate)
            at = at+1
            count = 0
            
        if(at >= len(averagingTimes)):
            break
    
    # Ensure that the eventRate output array has the same length as the average  weather data arrays.
    if( len(eventRate) < len(averagingTimes)):
        for i in range(len(averagingTimes-len(eventRate))):
            eventRate.append(0)
    
    return (eventRate)

#============================================================================================================================
#   The "Level 1" decomposition of the program consisting of sub-programs called directly from the Main program
#
#     dataInput()
#     processData():
#     outputData()


def dataInput():
    """
    Read the command line for details of files to be read,
    read the files and return their contents as PANDAS data frames.
    """
    (weatherFile, eventSummaryFile) = readCommandLine()	    # Obtain the names of the weather and event-summary files from
                                                            # the command line by invoking a function that does only this.
                                                            
    # For test purposes - hard-wire details of a couple of files with just use a short set of data.
    #weatherFile      = "../Nikhef501SummData/weather-s501-20190521.tsv"
    #eventSummaryFile = "../Nikhef501SummData/events-s501-20190521.tsv"
    
    if(debug >=1):
        print weatherFile,eventSummaryFile
    
    weatherData = readWeatherData(weatherFile)
    
    eventData   = readEventSummaryFile(eventSummaryFile)      # Get the data frame with the list of cosmic ray events.
    
    return (weatherData,eventData)


def processData(weatherData, eventData, averagingTime):
    """
    Turn the raw data read from HiSPARC into an output 
    Input:
        weatherData     a Python dictionary containing arrays of weather data.
                        The keys must include "times", "temperature", "pressure"
                        and the arrays associated with each of these keys must all be the same length.
        eventData       a Python Dictionary containing arrays of cosmic ray data.
                        The keys must include "times" and "eventRate", but will also include other stuff.
    
    Output:
        processedData   is returned and is a Python dictionary containing the averaged arrays (see below).
                        Each array will be of the same length.
    """
    
    (timeZero,weatherTimes,temperature,pressure) = aveWeatherData(averagingTime,weatherData) 
    
    averageEventRate = getAverageEventRate(timeZero,weatherTimes,eventData)   # Generate average rates over the intervals given by weatherTimes
    
    # Produce the output data as a Python dictionary
    processedData = {}
    
    processedData["times"]       = weatherTimes
    processedData["temperature"] = temperature
    processedData["pressure"]    = pressure
    processedData["eventRate"]   = averageEventRate
        
    return processedData


def outputData(processedData,csep):
    """
    Produce data in columns, separated by the character defined in the csep parameter.
    
    Input:
        processedData:  a Python dictionary containing a set of data arrays.
                        The data keys must correspond to those defined in processData().
                        All the arrays must have the same length.
    
        csep            The character used to separate data columns in the output file.
                        Typical choices are tab "\t" and comma "," both of which can be
                        readily handled by generic data analysis software such as Excel,
                        GNUPlot, etc.
    """
        
    times        = processedData["times"]
    temperatures = processedData["temperature"]
    pressures    = processedData["pressure"]   
    eventRate    = processedData["eventRate"]
    
    print len(times)
    print len(temperatures)
    print len(pressures)
    print len(eventRate)
    
    maxiter = min(len(times),len(temperatures),len(pressures),len(eventRate))
    
    print "year\tmonth\tday\thour\tminute\ttimestamp\ttemperature\tpressure\teventRate"
    for it in range(maxiter):
        aTime = datetime.datetime.utcfromtimestamp(times[it])           # Convert integer timestamp to Python dataTime
                                                                        # so that we can easily get year/month/day/hour.
        
        print aTime.year,csep, aTime.month,csep, aTime.day,csep, aTime.hour,csep,aTime.minute,csep,\
              times[it],csep,\
              temperatures[it],csep, pressures[it],csep, eventRate[it]
    
    plotScat("pressureVsTime",pressures,times,"Pressure","Time")
    plotScat("tempsVsTime",temperatures,times,"Temperatures","Time")
    plotScat("eventRatVsTime",eventRate,times,"Event Rate","Time")
    plotScat("eventRateVsPress",eventRate,pressures,"Event Rate","Pressure")
    plotScat("eventRateVsTemp",eventRate,temperatures,"Event Rate","Temperatures")    
    
    
def plotScat(outfilename, yval,xval,laby,labx):
        """Plot 
            outfilename     the root part of a file to which the graph is written.
                            The full name is <outfilename>_scatter.png .
            xval[]           A floating point list of .
            yval[]           A floating point list of .

        """
        fig, ax = plt.subplots(ncols=1, sharey=True, figsize=(18, 15))
        ax.scatter(xval,yval,s=1)
        #ax.axis([colmin, colmax, m606max, m606min])
        ax.set_title(outfilename)
        ax.set_xlabel(labx)
        ax.set_ylabel(laby)
    
        plt.savefig(outfilename+".png")
        if(debug >= 1):
            print "   Generated scatter plot to file ",outfilename+"_scatter.png"
#============================================================================================================================
    
if __name__ == "__main__":
    """
    Divide the handling of HiSPARC data into three separate phases: input, processing, output.
    Each phase is completed before the next starts.
    
    Note that at this level of the program we defer describing exactly what we are storing in
    our packages of data (weatherData, eventData), nor do we need to specify here how it is organised in these packages.
    (This gives us flexibility to easily change the organisation and possibly the contents of these data packages 
    at a later time. We might, for example, switch to using the flexible PANDAS library for packaging data.)
    """
    
    (weatherData, eventData) = dataInput()      # At this point it is not necessary to specify exactly how the 
                                                # weather and event data are organised.
                                                # We just need to know that everthing that may be required
                                                # will be passed in the weatherData and eventData information structures.
    
    processedData = processData(weatherData, eventData,3600)   # Everything will be passed for processing to this sub-program
                                                               # Once again, we do not need to specify at this point the
                                                               # exact contents and organisation of information in processedData.
                                                               # This gives flexibility to change these structures later
                                                               # without the need to adjust this part of the code.
    
    outputData(processedData,"\t")              # Produce an output file consisting of tab-separated columns of data.
